#!/bin/sh
# tunelfso.sh -- modify relocations of shared library to speedup exec
# by reducing symbol searches when library is fetched and relocated.
# $Id: tunelfso.sh,v 1.3 1999/09/24 21:27:12 jreiser Exp $

# Copyright 1999 John F. Reiser.  All rights reserved.
# Contact: jreiser@BitWagon.com
# This file may be used, modified, and distributed under the terms
# of the GNU General Public License, version 2.

# Demo of problem: see stderr of /bin/sh command "LD_DEBUG=symbols /bin/date".
# About 860 runtime symbol lookups are required to run this small program!
# Commonly tuned modules:
#	/lib/ld-linux.so.2  ->  ld-2.1.1.so
#	/lib/libc.so.6  ->  libc-2.1.1.so
# After tuning these two files, about 440 lookups remain (a savings of 420).
#
# WARNING:  REPLACING THESE FILES IS DANGEROUS.  IF ANYTHING GOES WRONG,
# THEN THE SYSTEM MIGHT NOT BOOT AT ALL.  DO NOT REPLACE THESE FILES
# UNLESS YOU HAVE A _TESTED_ AND _SUCCESSFUL_ ALTERNATE BOOT MECHANISM,
# SUCH AS A SECOND ROOT PARTITION, OR A BOOT+ROOT "RESCUE" FLOPPY DISK.
# TAKE A BACKUP OF THE ROOT PARTITION BEFORE YOU BEGIN.  YOU HAVE BEEN WARNED.
# -----


# Exit immediately if an error occurs.
set -e

# First parameter is the library to modify; default is libc.
# 'shift' complains if this script is invoked with no parameters;
# inhibit error exit for this case.
LIB=${1:-/lib/libc.so.6};  shift  2>/dev/null  ||  true

# Remaining parameters are the list of applications; default is netscape.
APPS=${@:-/usr/lib/netscape/netscape-navigator}

# Start with list of symbols that always must be allowed to be overridden.
# 1st group (_IO_std etc.) is required when tuning /lib/ld-linux.so;
# else booting fails, with /sbin/init complaining about respawning too fast.
# 2nd group (malloc etc.) is commonly overridden using LD_PRELOAD.
echo '
^_IO_std
^_IO_getc$
^_IO_putc$
^_environ$
^__environ$

^malloc$
^calloc$
^realloc$
^free$
^cfree$
^memalign$
'  > apps.overrides

# Get the list of shared libraries used by the $APPS, excluding $LIB.
ldd $APPS |
grep -v $LIB |
sed -e 's/.*=>//' -e 's/(.*//' |
sort -u  > apps.modules   # remove duplicates for speed, sort for clarity

# Get the sorted list of symbols defined by the $APPS and shared libraries.
nm --dynamic $APPS `cat apps.modules` |
grep -v ' U ' |
sed -e 's/.* //' |
sort -u  > apps.defs   # sorted list of unique symbols is required by 'comm'

# rm apps.modules   # Done with this file; kept for audit.

# Get the sorted list of symbols defined by the $LIB,
# intersect with the sorted list of symbols defined by the $APPS,
# and construct regular expression to match each whole symbol.
# These are the symbols in $LIB that are overridden by symbols in the $APPS.
nm --dynamic $LIB  |
grep -v ' U ' |
sed -e 's/.* //' |
comm -12 - apps.defs |
sed -e 's/.*/^&$/'  >> apps.overrides

# rm apps.defs   # Done with this file; kept for audit.

# Change the DYNAMIC relocations of $LIB to avoid runtime symbol searches
# for symbols which begin with '_'  [ -i'^_' ],
# except for symbols which are overridden  [ -x@apps.overrides ].
# Send to stdout the sorted list of symbols whose relocations are modified.
# The new library is in `basename $LIB`, which is a BUG if $LIB has no
# path for 'basename' to remove (output is created before input is read).
./tunelfso  -i'^_'  -x@apps.overrides  $LIB  2>&1  1>`basename $LIB`  |
sort -u

# rm apps.overrides   # Done with this file; kept for audit.

# end-of-file
